(function() {
  "use strict";

  // Switch on logging by default.
  process.env.DEBUG = "*";

    // Load the nqm input module for receiving input from the process host.
  var input = require("nqm-process-utils").input;

  // Load the nqm output module for communicating progress and info back to the process host.
  var output = require("nqm-process-utils").output;

  var path = require("path");
  var fs = require("fs");  
  var args = require("minimist")(process.argv.slice(2));
  var util = require("util");
  var Promise = require("bluebird");
  var _ = require("lodash");
  var moment = require("moment");
  var TDXapi = require("nqm-api-tdx");
  var reformat = require("./lib/reformat");
  var convertor = require("./lib/event-store-convertor");
  var checkNewLine = require("./lib/check-newline");
  var tdxApi = null;
  var config = null;
  var configFile = args.config || "./config.json";
  var _sourceFilePreparationProgress;   // Estimate source file preparation will take 25% of total time.
  var _progress = 0;
  var _progressStep = 3; 

  var abort = function() {
    var msg = util.format.apply(util, arguments);
    output.debug("aborting - " + msg);
    output.error(msg);
    process.exit(1);
  };

  var loadArguments = function() {
    try {
      config = require(configFile);
    } catch(e) {
      abort("failed to load config file: " + e.message);
    }

    config.readHost = args.readHost || config.readHost || "localhost:27017";
    config.esHost = args.esHost || config.esHost || "localhost:27017";
    config.commandHost = args.commandHost || config.commandHost || "https://cmd.nqminds.com";
    config.queryHost = args.queryHost || config.queryHost || "https://q.nqminds.com";
    config.copyNameIndex = 0;
  };

  var checkArguments = function() {
    /*
     * Check system (i.e. non-task) arguments.
     * These should be configured when the process is installed.
     */
    if (!config.readDb) {
      abort("no read db connection string");
    }
    if (!config.esDb) {
      abort("no event store db connection string");
    }
    if (!config.commandHost) {
      abort("no command host given");
    }
    if (!config.queryHost) {
      abort("no query host given");
    }
    if (!config.resourceStorePath) {
      abort("no resource store path");
    }

    /*
     * Check task-instance arguments.
     * These are specified by the user when the task is started.
     */
    if (!config.sourceResource) {
      abort("no source resource given");
    }
    if (!config.accessToken) {
      abort("no accessToken given");
    }
    if (!config.targetDataset) {
      if (!config.targetSchema) {
        abort("no targetSchema given");
      }
      if (!config.targetFolder) {
        abort("no targetFolder given");
      }
      if (!config.targetName) {
        config.targetName = config.targetSchema + " import " + moment().format("Do MMM YYYY");
      }
    }
  };

  var initialiseTDXApi = function() {
    tdxApi = new TDXapi(config);
    Promise.promisifyAll(tdxApi);
    updateProgress();
    return Promise.resolve();
  };

  var getSourceFile = function() {
    config.sourceFile = path.join(config.resourceStorePath, config.sourceResource + ".json");
    try {
      // Check file is accessible.
      fs.accessSync(config.sourceFile, fs.F_OK);
      updateProgress();
    } catch (e) {
      output.debug("can't access source file " + config.sourceFile + " error is: " + e.message);
      abort("sourceResource input '%s' is invalid - it must be a TDX resource of type rawFile/json with data uploaded.",config.sourceResource);
    }
  };

  var createDataset = function() {
    var datasetPromise;

    if (!config.targetDataset) {
      var targetName = config.copyNameIndex > 0 ? config.targetName + " copy " + config.copyNameIndex : config.targetName;
      // Create the dataset with 'importing' flag set so that the data projection does not start until the import is complete.
      datasetPromise = tdxApi.createDatasetAsync({ name: targetName, basedOnSchema: config.targetSchema, parentId: config.targetFolder, importing: true })
        .then(function(createResult) {
          output.debug("created dataset id: %j", createResult);
          config.targetDataset = createResult.response.id;

          updateProgress();
        })
        .catch(function(err) {
          // TODO - add error codes to the entire TDX!
          if (err.message.indexOf("resource already exists with name:") >= 0) {
            output.debug("name clash - bumping number and trying again");
            config.copyNameIndex++;
            return createDataset();
          } else {
            abort("unable to create target dataset '%s' using schema '%s' in folder '%s' [%s]", targetName, config.targetSchema, config.targetFolder, err.message);
          }
        });
    } else {
      // A target dataset was given - need to reset the datastore?
      // -OR-
      // Just switch off data projection?
      output.debug("importing to existing dataset - switching off projection");
      updateProgress();
      datasetPromise = setDatasetImporting(true);
    }
    
    return datasetPromise
      .then(function() {
        // Delay here to give the projection time to catch up.
        return new Promise(function(resolve, reject) {
          setTimeout(function() {
            updateProgress();
            resolve(config.targetDataset);
          }, 10000);
        });        
      });
  };

  var getTargetDataset = function() {
    return tdxApi.queryAsync("datasets/" + config.targetDataset, null, null, null)
      .then(function(dataset) {
        output.debug("retrieved dataset %j", dataset);
        config.dataset = dataset;

        if (!config.targetName) {
          config.targetName = dataset.name;
        }

        updateProgress();
        return Promise.resolve(dataset);
      })
      .catch(function(err) {
        abort("failed to get target dataset '%s' [%s]", config.targetDataset, err.message);
      });
  };

  var setDatasetImporting = function(importing) {
    if (importing) {
      output.debug("importing flag set - stopping data projection");
    } else {
      output.debug("importing flag cleared - starting data projection");
    }
    // Clear the dataset 'importing' flag so that the data projection can start.
    return tdxApi.setDatasetImportFlagAsync(config.targetDataset, importing)
      .then(function(result) {
        output.debug("setDatasetImporting to '%s' result: %j", importing, result);
        return Promise.resolve();
      })
      .catch(function(err) {
        abort("failed to start data projection for dataset '%s' [%s]", config.targetDataset, err.message);
      });
  };

  var doImport = function(importIdx, dbHost, dbName, collectionName, sourceFile) {
    var spawn = require("child_process").spawn;

    var totalProgress = (100-_sourceFilePreparationProgress)/2;

    output.debug("starting import of %s into %s/%s/%s", sourceFile, dbHost, dbName, collectionName);
    return new Promise(function(resolve, reject) {
      var importArgs = ["--host",dbHost,"--db",dbName,"--collection", collectionName,"--file", sourceFile];
      if (config.truncate) {
        importArgs.push("--drop");
      }
      var importChild = spawn("mongoimport", importArgs);
      
      // Intercept mongoimport STDOUT - attempt to parse the progress string. 
      importChild.stdout.on("data", function(stdout) {
        stdout = stdout.toString();
        output.debug("import - %s", stdout);
        var progressLabel = "%";
        var progressIdx = stdout.indexOf(progressLabel); 
        if (progressIdx >= 0) {
          var progress = parseInt(stdout.substr(progressIdx - 2,3).trim());
          
          // Adjust progress by _sourceFilePreparationProgress and the fact there are two imports to run.
          var adjustedProgress = _sourceFilePreparationProgress + totalProgress*importIdx + totalProgress*(progress/100);
          output.progress(adjustedProgress);
        }      
      });

      importChild.on("close", function(code) {
        output.debug("exited with code %d", code);
        if (code) {
          abort("Import failed with exit code %d", code);
        } else {
          // Finised with zero exit code - update final progress.
          output.progress(_sourceFilePreparationProgress + totalProgress*(importIdx+1));
          resolve();
        }
      });

      importChild.on("error", function(err) {
        output.error("importer error: " + err.message);
      });
    });
  };

  var runImport = function() {
    output.debug("runImport");

    _sourceFilePreparationProgress = _progress;

    var dataset = config.dataset;

    // First import the event store collection.
    return doImport(0, config.esHost, config.esDb, dataset.store + ".events", config.eventSourceFile)
      .then(function() {
        // Then import the read-side collection.
        return doImport(1, config.readHost, config.readDb, dataset.store, config.readSourceFile);
      })
      .catch(function(err) {
        abort("unexpected error during core import: %s", err.message);
      });
  };

  var prepareSourceFiles = function() {
    output.debug("preparing source files for import");

    return checkNewLine(config.sourceFile, output)
      .then(function() {
        updateProgress();
        config.eventSourceFile = config.sourceFile + ".es.json";
        if (config.sourceDataPath) {
          // A data path is given, so assume the files need re-formatting from proper JSON 
          // with sourceDataPath pointing to the array to import, rather than new-line separated JSON documents
          // as required by mongoimport.
          output.debug("have sourceDataPath => re-formating for both read-side and event store");
          config.readSourceFile = config.sourceFile + ".read.json";
          return reformat(config.sourceDataPath + ".*", config.sourceFile, config.readSourceFile, config.eventSourceFile, output)
            .then(function() {
              updateProgress();
            })
            .catch(function(err) {
              output.debug(err.stack);          
              abort("failure during source file re-formatting: %s", err.message);
            });
        } else {
          // No data path => just need to create the event-store version of the source file.
          output.debug("no sourceDataPath => generating event store source file");
          config.readSourceFile = config.sourceFile;
          return convertor(config.sourceFile, config.eventSourceFile, output)
            .then(function() {
              updateProgress();
            })
            .catch(function(err) {
              output.debug(err.stack);          
              abort("failure during conversion to event-store format: %s", err.message);
            });
        }
      });
  };

  var updateProgress = function() {
    _progress += _progressStep;
    output.progress(_progress);
  };

  loadArguments();

  input.read(function(err, input) {
    _.extend(config,input.data);

    output.progress(0);

    config.accessToken = input.authToken;
    checkArguments();  

    initialiseTDXApi()
    .then(getSourceFile)
    .then(createDataset)
    .then(getTargetDataset)
    .then(prepareSourceFiles)
    .then(runImport)
    .then(setDatasetImporting)
    .then(function() {
      output.debug("finished");
      output.result(JSON.stringify({ok: true}));
      output.progress(100);        
    })
    .catch(function(err) {
      output.debug("uncaught exception [%s]", err.message);
      output.debug(err.stack);
      process.exit(1);
    });
  });
}());